module Cidr
( Cidr(..),
  cidr_from_string,
  cidr_tests,
  combine_all,
  contains,
  contains_proper,
  max_octet1,
  max_octet2,
  max_octet3,
  max_octet4,
  min_octet1,
  min_octet2,
  min_octet3,
  min_octet4,
  prop_all_cidrs_contain_themselves,
  prop_contains_proper_intransitive
) where

import Data.List (nubBy)
import Test.HUnit
import Test.QuickCheck

import qualified Bit as B
import IPv4Address
import ListUtils
import Maskable
import Maskbits
import Octet


data Cidr = None | Cidr { ipv4address :: IPv4Address,
                          maskbits :: Maskbits }
            deriving (Eq)


instance Show Cidr where
    show Cidr.None = "None"
    show cidr = (show (ipv4address cidr)) ++ "/" ++ (show (maskbits cidr))


instance Arbitrary Cidr where
    arbitrary = do
      ipv4 <- arbitrary :: Gen IPv4Address
      mask <- arbitrary :: Gen Maskbits
      return (Cidr ipv4 mask)

    coarbitrary _ = variant 0


-- Two CIDR ranges are equivalent if they have the same network bits
-- and the masks are the same.
equivalent :: Cidr -> Cidr -> Bool
equivalent Cidr.None Cidr.None = True
equivalent Cidr.None _ = False
equivalent _ Cidr.None = False
equivalent (Cidr addr1 mbits1) (Cidr addr2 mbits2) =
    (mbits1 == mbits2) && ((apply_mask addr1 mbits1) == (apply_mask addr2 mbits2))

-- Returns the mask portion of a CIDR address. That is, everything
-- after the trailing slash.
maskbits_from_cidr_string :: String -> Maskbits
maskbits_from_cidr_string s =
    maskbits_from_string ((splitWith (`elem` "/") s) !! 1)


-- Takes an IP address String in CIDR notation, and returns a list of
-- its octets (as Ints).
octets_from_cidr_string :: String -> [Octet]
octets_from_cidr_string s =
    map octet_from_string (take 4 (splitWith (`elem` "./") s))


cidr_from_string :: String -> Cidr
cidr_from_string s
    | addr  == IPv4Address.None = Cidr.None
    | mbits == Maskbits.None    = Cidr.None
    | otherwise = Cidr addr mbits
               where
                 addr = ipv4address_from_octets (oct1) (oct2) (oct3) (oct4)
                 oct1 = (octs !! 0)
                 oct2 = (octs !! 1)
                 oct3 = (octs !! 2)
                 oct4 = (octs !! 3)
                 octs  = octets_from_cidr_string s
                 mbits = maskbits_from_cidr_string s



min_host :: Cidr -> IPv4Address
min_host Cidr.None                 = IPv4Address.None
min_host (Cidr IPv4Address.None _) = IPv4Address.None
min_host (Cidr _ Maskbits.None)    = IPv4Address.None
min_host (Cidr addr mask)
    | mask == ThirtyTwo = addr
    | mask == ThirtyOne = addr { octet4 = apply_mask oct4 Seven }
    | mask == Thirty = addr { octet4 = apply_mask oct4 Six }
    | mask == TwentyNine = addr { octet4 = apply_mask oct4 Five }
    | mask == TwentyEight = addr { octet4 = apply_mask oct4 Four }
    | mask == TwentySeven = addr { octet4 = apply_mask oct4 Three }
    | mask == TwentySix = addr { octet4 = apply_mask oct4 Two }
    | mask == TwentyFive = addr { octet4 = apply_mask oct4 One }
    | mask == TwentyFour = addr { octet4 = min_octet }
    | mask == TwentyThree = addr { octet3 = apply_mask oct3 Seven,
                                   octet4 = min_octet }
    | mask == TwentyTwo = addr { octet3 = apply_mask oct3 Six,
                                 octet4 = min_octet }
    | mask == TwentyOne = addr { octet3 = apply_mask oct3 Five,
                                 octet4 = min_octet }
    | mask == Twenty = addr { octet3 = apply_mask oct3 Four,
                              octet4 = min_octet }
    | mask == Nineteen = addr { octet3 = apply_mask oct3 Three,
                                octet4 = min_octet }
    | mask == Eighteen = addr { octet3 = apply_mask oct3 Two,
                                octet4 = min_octet }
    | mask == Seventeen = addr { octet3 = apply_mask oct3 One,
                                 octet4 = min_octet }
    | mask == Sixteen = addr { octet3 = min_octet,
                               octet4 = min_octet }
    | mask == Fifteen = addr { octet2 = apply_mask oct2 Seven,
                               octet3 = min_octet,
                               octet4 = min_octet }
    | mask == Fourteen = addr { octet2 = apply_mask oct2 Six,
                                octet3 = min_octet,
                                octet4 = min_octet }
    | mask == Thirteen = addr { octet2 = apply_mask oct2 Five,
                                octet3 = min_octet,
                                octet4 = min_octet }
    | mask == Twelve = addr { octet2 = apply_mask oct2 Four,
                              octet3 = min_octet,
                              octet4 = min_octet }
    | mask == Eleven = addr { octet2 = apply_mask oct2 Three,
                              octet3 = min_octet,
                              octet4 = min_octet }
    | mask == Ten = addr { octet2 = apply_mask oct2 Two,
                           octet3 = min_octet,
                           octet4 = min_octet }
    | mask == Nine = addr { octet2 = apply_mask oct2 One,
                            octet3 = min_octet,
                            octet4 = min_octet }
    | mask == Eight = min_address { octet1 = oct1 }
    | mask == Seven = min_address { octet1 = apply_mask oct1 Seven }
    | mask == Six = min_address { octet1 = apply_mask oct1 Six }
    | mask == Five = min_address { octet1 = apply_mask oct1 Five }
    | mask == Four = min_address { octet1 = apply_mask oct1 Four }
    | mask == Three = min_address { octet1 = apply_mask oct1 Three }
    | mask == Two = min_address { octet1 = apply_mask oct1 Two }
    | mask == One = min_address { octet1 = apply_mask oct1 One }
    | mask == Zero = min_address
    | otherwise = addr
    where
      oct1 = (octet1 addr)
      oct2 = (octet2 addr)
      oct3 = (octet3 addr)
      oct4 = (octet4 addr)



max_host :: Cidr -> IPv4Address
max_host Cidr.None                 = IPv4Address.None
max_host (Cidr IPv4Address.None _) = IPv4Address.None
max_host (Cidr _ Maskbits.None)    = IPv4Address.None
max_host (Cidr addr mask)
    | mask == ThirtyTwo = addr
    | mask == ThirtyOne = addr { octet4 = oct4 { b8 = B.One } }
    | mask == Thirty = addr { octet4 = oct4 { b7 = B.One, b8 = B.One } }
    | mask == TwentyNine = addr { octet4 = oct4 { b6 = B.One,
                                                  b7 = B.One,
                                                  b8 = B.One } }
    | mask == TwentyEight = addr { octet4 = oct4 { b5 = B.One,
                                                   b6 = B.One,
                                                   b7 = B.One,
                                                   b8 = B.One } }
    | mask == TwentySeven = addr { octet4 = oct4 { b4 = B.One,
                                                   b5 = B.One,
                                                   b6 = B.One,
                                                   b7 = B.One,
                                                   b8 = B.One } }
    | mask == TwentySix = addr { octet4 = oct4 { b3 = B.One,
                                                 b4 = B.One,
                                                 b5 = B.One,
                                                 b6 = B.One,
                                                 b7 = B.One,
                                                 b8 = B.One } }
    | mask == TwentyFive = addr { octet4 = oct4 { b2 = B.One,
                                                  b3 = B.One,
                                                  b4 = B.One,
                                                  b5 = B.One,
                                                  b6 = B.One,
                                                  b7 = B.One,
                                                  b8 = B.One } }
    | mask == TwentyFour = addr { octet4 = max_octet }
    | mask == TwentyThree = ipv4address_from_octets oct1 oct2 (Octet a17 a18 a19 a20 a21 a22 a23 B.One) (max_octet)
    | mask == TwentyTwo = ipv4address_from_octets oct1 oct2 (Octet a17 a18 a19 a20 a21 a22 B.One B.One) (max_octet)
    | mask == TwentyOne = ipv4address_from_octets oct1 oct2 (Octet a17 a18 a19 a20 a21 B.One B.One B.One) (max_octet)
    | mask == Twenty = ipv4address_from_octets oct1 oct2 (Octet a17 a18 a19 a20 B.One B.One B.One B.One) (max_octet)
    | mask == Nineteen = ipv4address_from_octets oct1 oct2 (Octet a17 a18 a19 B.One B.One B.One B.One B.One) (max_octet)
    | mask == Eighteen = ipv4address_from_octets oct1 oct2 (Octet a17 a18 B.One B.One B.One B.One B.One B.One) (max_octet)
    | mask == Seventeen = ipv4address_from_octets oct1 oct2 (Octet a17 B.One B.One B.One B.One B.One B.One B.One) (max_octet)
    | mask == Sixteen = ipv4address_from_octets oct1 oct2 (max_octet) (max_octet)
    | mask == Fifteen = ipv4address_from_octets oct1 (Octet a9 a10 a11 a12 a13 a14 a15 B.One) (max_octet) (max_octet)
    | mask == Fourteen = ipv4address_from_octets oct1 (Octet a9 a10 a11 a12 a13 a14 B.One B.One) (max_octet) (max_octet)
    | mask == Thirteen = ipv4address_from_octets oct1 (Octet a9 a10 a11 a12 a13 B.One B.One B.One) (max_octet) (max_octet)
    | mask == Twelve = ipv4address_from_octets oct1 (Octet a9 a10 a11 a12 B.One B.One B.One B.One) (max_octet) (max_octet)
    | mask == Eleven = ipv4address_from_octets oct1 (Octet a9 a10 a11 B.One B.One B.One B.One B.One) (max_octet) (max_octet)
    | mask == Ten = ipv4address_from_octets oct1 (Octet a9 a10 B.One B.One B.One B.One B.One B.One) (max_octet) (max_octet)
    | mask == Nine = ipv4address_from_octets oct1 (Octet a9 B.One B.One B.One B.One B.One B.One B.One) (max_octet) (max_octet)
    | mask == Eight = ipv4address_from_octets oct1 (max_octet) (max_octet) (max_octet)
    | mask == Seven = ipv4address_from_octets (Octet a1 a2 a3 a4 a5 a6 a7 B.One) (max_octet) (max_octet) (max_octet)
    | mask == Six = ipv4address_from_octets (Octet a1 a2 a3 a4 a5 a6 B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == Five = ipv4address_from_octets (Octet a1 a2 a3 a4 a5 B.One B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == Four = ipv4address_from_octets (Octet a1 a2 a3 a4 B.One B.One B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == Three = ipv4address_from_octets (Octet a1 a2 a3 B.One B.One B.One B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == Two = ipv4address_from_octets (Octet a1 a2 B.One B.One B.One B.One B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == One = ipv4address_from_octets (Octet a1 B.One B.One B.One B.One B.One B.One B.One) (max_octet) (max_octet) (max_octet)
    | mask == Zero = ipv4address_from_octets (max_octet) (max_octet) (max_octet) (max_octet)
    | otherwise = addr
    where
      a1  = (b1 oct1)
      a2  = (b2 oct1)
      a3  = (b3 oct1)
      a4  = (b4 oct1)
      a5  = (b5 oct1)
      a6  = (b6 oct1)
      a7  = (b7 oct1)
      a9  = (b1 oct2)
      a10 = (b2 oct2)
      a11 = (b3 oct2)
      a12 = (b4 oct2)
      a13 = (b5 oct2)
      a14 = (b6 oct2)
      a15 = (b7 oct2)
      a17 = (b1 oct3)
      a18 = (b2 oct3)
      a19 = (b3 oct3)
      a20 = (b4 oct3)
      a21 = (b5 oct3)
      a22 = (b6 oct3)
      a23 = (b7 oct3)
      oct1 = (octet1 addr)
      oct2 = (octet2 addr)
      oct3 = (octet3 addr)
      oct4 = (octet4 addr)



min_octet1 :: Cidr -> Octet
min_octet1 cidr = octet1 (min_host cidr)

min_octet2 :: Cidr -> Octet
min_octet2 cidr = octet2 (min_host cidr)

min_octet3 :: Cidr -> Octet
min_octet3 cidr = octet3 (min_host cidr)

min_octet4 :: Cidr -> Octet
min_octet4 cidr = octet4 (min_host cidr)

max_octet1 :: Cidr -> Octet
max_octet1 cidr = octet1 (max_host cidr)

max_octet2 :: Cidr -> Octet
max_octet2 cidr = octet2 (max_host cidr)

max_octet3 :: Cidr -> Octet
max_octet3 cidr = octet3 (max_host cidr)

max_octet4 :: Cidr -> Octet
max_octet4 cidr = octet4 (max_host cidr)



-- Return true if the first argument (a CIDR range) contains the
-- second (another CIDR range). There are a lot of ways we can be fed
-- junk here. For lack of a better alternative, just return False when
-- we are given nonsense.
contains :: Cidr -> Cidr -> Bool
contains Cidr.None _ = False
contains _ Cidr.None = False
contains (Cidr _ Maskbits.None) _    = False
contains (Cidr IPv4Address.None _) _ = False
contains _ (Cidr _ Maskbits.None)    = False
contains _ (Cidr IPv4Address.None _) = False

-- If the number of bits in the network part of the first address is
-- larger than the number of bits in the second, there is no way that
-- the first range can contain the second. For, if the number of
-- network bits is larger, then the number of host bits must be
-- smaller, and if cidr1 has fewer hosts than cidr2, cidr1 most
-- certainly does not contain cidr2.
--
-- On the other hand, if the first argument (cidr1) has fewer (or the
-- same number of) network bits as the second, it can contain the
-- second. In this case, we need to check that every host in cidr2 is
-- contained in cidr1. If a host in cidr2 is contained in cidr1, then
-- at least mbits1 of an address in cidr2 will match cidr1. For
-- example,
--
-- cidr1 = 192.168.1.0/23, cidr2 = 192.168.1.100/24
--
-- Here, cidr2 contains all of 192.168.1.0 through
-- 192.168.1.255. However, cidr1 contains BOTH 192.168.0.0 through
-- 192.168.0.255 and 192.168.1.0 through 192.168.1.255. In essence,
-- what we want to check is that cidr2 "begins with" something that
-- cidr1 CAN begin with. Since cidr1 can begin with 192.168.1, and
-- cidr2 DOES, cidr1 contains cidr2..
--
-- The way that we check this is to apply cidr1's mask to cidr2's
-- address and see if the result is the same as cidr1's mask applied
-- to cidr1's address.
--
contains (Cidr addr1 mbits1) (Cidr addr2 mbits2)
    | mbits1 > mbits2 = False
    | otherwise = addr1masked == addr2masked
    where
      addr1masked = apply_mask addr1 mbits1
      addr2masked = apply_mask addr2 mbits1


contains_proper :: Cidr -> Cidr -> Bool
contains_proper cidr1 cidr2 =
    (cidr1 `contains` cidr2) && (not (cidr2 `contains` cidr1))


-- A CIDR range is redundant (with respect to the given list) if
-- another CIDR range in that list properly contains it.
redundant :: [Cidr] -> Cidr -> Bool
redundant cidrlist cidr = any ((flip contains_proper) cidr) cidrlist


-- First, we look at all possible pairs of cidrs, and combine the
-- adjacent ones in to a new list. Then, we concatenate that list with
-- the original one, and filter out all of the redundancies. If two
-- adjacent Cidrs are combined into a larger one, they will be removed
-- in the second step since the larger Cidr must contain the smaller
-- two.
combine_all :: [Cidr] -> [Cidr]
combine_all cidrs =
    combine_contained unique_cidrs
    where
      unique_cidrs = nubBy equivalent valid_cidr_combinations
      valid_cidr_combinations = filter (/= Cidr.None) cidr_combinations
      cidr_combinations =
          cidrs ++ [ (combine_adjacent x y) | x <- cidrs, y <- cidrs ]


-- Take a list of CIDR ranges and filter out all of the ones that are
-- contained entirelt within some other range in the list.
combine_contained :: [Cidr] -> [Cidr]
combine_contained cidrs =
    filter (not . (redundant cidrs)) cidrs


-- If the two Cidrs are not adjacent, return Cidr.None. Otherwise,
-- decrement the maskbits of cidr1 and return that; it will contain
-- both cidr1 and cidr2.
combine_adjacent :: Cidr -> Cidr -> Cidr
combine_adjacent cidr1 cidr2
  | not (adjacent cidr1 cidr2) = Cidr.None
  | (maskbits cidr1 == Zero) = Cidr.None
  | otherwise = cidr1 { maskbits = decrement (maskbits cidr1) }



-- Determine whether or not two CIDR ranges are adjacent. If two
-- ranges lie consecutively within the IP space, they can be
-- combined. For example, 10.1.0.0/24 and 10.0.1.0/24 are adjacent,
-- and can be combined in to 10.1.0.0/23.
adjacent :: Cidr -> Cidr -> Bool
adjacent Cidr.None _ = False
adjacent _ Cidr.None = False
adjacent cidr1 cidr2
  | mbits1 /= mbits2 = False
  | mbits1 == Maskbits.Zero = False -- They're equal.
  | otherwise = (mbits1 == (most_sig_bit_different addr1 addr2))
  where
    addr1 = ipv4address cidr1
    addr2 = ipv4address cidr2
    mbits1 = maskbits cidr1
    mbits2 = maskbits cidr2





-- HUnit Tests

test_equality1 :: Test
test_equality1 =
    TestCase $ assertEqual "10.1.1.0/23 equals itself" True (cidr1 == cidr1)
    where
      cidr1 = cidr_from_string "10.1.1.0/23"


test_contains1 :: Test
test_contains1 =
    TestCase $ assertEqual "10.1.1.0/23 contains 10.1.1.0/24" True (cidr1 `contains` cidr2)
    where
      cidr1 = cidr_from_string "10.1.1.0/23"
      cidr2 = cidr_from_string "10.1.1.0/24"


test_contains2 :: Test
test_contains2 =
    TestCase $ assertEqual "10.1.1.0/23 contains itself" True (cidr1 `contains` cidr1)
    where
      cidr1 = cidr_from_string "10.1.1.0/23"


test_contains_proper1 :: Test
test_contains_proper1 =
    TestCase $ assertEqual "10.1.1.0/23 contains 10.1.1.0/24 properly" True (cidr1 `contains_proper` cidr2)
    where
      cidr1 = cidr_from_string "10.1.1.0/23"
      cidr2 = cidr_from_string "10.1.1.0/24"


test_contains_proper2 :: Test
test_contains_proper2 =
    TestCase $ assertEqual "10.1.1.0/23 does not contain itself properly" False (cidr1 `contains_proper` cidr1)
    where
      cidr1 = cidr_from_string "10.1.1.0/23"


test_adjacent1 :: Test
test_adjacent1 =
    TestCase $ assertEqual "10.1.0.0/24 is adjacent to 10.1.1.0/24" True (cidr1 `adjacent` cidr2)
    where
      cidr1 = cidr_from_string "10.1.0.0/24"
      cidr2 = cidr_from_string "10.1.1.0/24"


test_adjacent2 :: Test
test_adjacent2 =
    TestCase $ assertEqual "10.1.0.0/23 is not adjacent to 10.1.0.0/24" False (cidr1 `adjacent` cidr2)
    where
      cidr1 = cidr_from_string "10.1.0.0/23"
      cidr2 = cidr_from_string "10.1.0.0/24"


test_adjacent3 :: Test
test_adjacent3 =
    TestCase $ assertEqual "10.1.0.0/24 is not adjacent to 10.2.5.0/24" False (cidr1 `adjacent` cidr2)
    where
      cidr1 = cidr_from_string "10.1.0.0/24"
      cidr2 = cidr_from_string "10.2.5.0/24"


test_adjacent4 :: Test
test_adjacent4 =
    TestCase $ assertEqual "10.1.1.0/24 is not adjacent to 10.1.2.0/24" False (cidr1 `adjacent` cidr2)
    where
      cidr1 = cidr_from_string "10.1.1.0/24"
      cidr2 = cidr_from_string "10.1.2.0/24"


test_combine_contained1 :: Test
test_combine_contained1 =
    TestCase $ assertEqual "10.0.0.0/8, 10.1.0.0/16, and 10.1.1.0/24 combine to 10.0.0.0/8" expected_cidrs (combine_contained test_cidrs)
    where
      cidr1 = cidr_from_string "10.0.0.0/8"
      cidr2 = cidr_from_string "10.1.0.0/16"
      cidr3 = cidr_from_string "10.1.1.0/24"
      expected_cidrs = [cidr1]
      test_cidrs = [cidr1, cidr2, cidr3]


test_combine_contained2 :: Test
test_combine_contained2 =
    TestCase $ assertEqual "192.168.3.0/23 does not contain 192.168.1.0/24" [cidr1, cidr2] (combine_contained [cidr1, cidr2])
    where
      cidr1 = cidr_from_string "192.168.3.0/23"
      cidr2 = cidr_from_string "192.168.1.0/24"


test_combine_all1 :: Test
test_combine_all1 =
    TestCase $ assertEqual "10.0.0.0/24 is adjacent to 10.0.1.0/24 and 10.0.3.0/23 contains 10.0.2.0/24" expected_cidrs (combine_all test_cidrs)
    where
      cidr1 = cidr_from_string "10.0.0.0/24"
      cidr2 = cidr_from_string "10.0.1.0/24"
      cidr3 = cidr_from_string "10.0.2.0/24"
      cidr4 = cidr_from_string "10.0.3.0/23"
      cidr5 = cidr_from_string "10.0.0.0/23"
      expected_cidrs = [cidr4, cidr5]
      test_cidrs = [cidr1, cidr2, cidr3, cidr4]


test_combine_all2 :: Test
test_combine_all2 =
    TestCase $ assertEqual "127.0.0.1/32 combines with itself recursively" expected_cidrs (combine_all test_cidrs)
    where
      cidr1 = cidr_from_string "127.0.0.1/32"
      expected_cidrs = [cidr1]
      test_cidrs = [cidr1, cidr1, cidr1, cidr1, cidr1]


cidr_tests :: [Test]
cidr_tests = [ test_equality1,
               test_contains1,
               test_contains2,
               test_contains_proper1,
               test_contains_proper2,
               test_adjacent1,
               test_adjacent2,
               test_adjacent3,
               test_adjacent4,
               test_combine_contained1,
               test_combine_contained2,
               test_combine_all1,
               test_combine_all2
             ]


-- QuickCheck Tests
prop_all_cidrs_contain_themselves :: Cidr -> Bool
prop_all_cidrs_contain_themselves cidr1 = cidr1 `contains` cidr1


-- If cidr1 properly contains cidr2, then by definition cidr2
-- does not properly contain cidr1.
prop_contains_proper_intransitive :: Cidr -> Cidr -> Property
prop_contains_proper_intransitive cidr1 cidr2 =
    (cidr1 `contains_proper` cidr2) ==>
      (not (cidr2 `contains_proper` cidr1))
